!-----------------------------------------------------------------------------
! Magic And Fight System --- SPELL SYSTEM
! For INFORM compiler and Z-Machine text based Adventure games
!-----------------------------------------------------------------------------
!  Author: Giancarlo Niccolai
!  Version: $Id: MFS_Magic.h,v 1.9 2003/11/18 07:59:27 gian Exp $
!  Contact author at: giancarlo@niccolai.ws
!  Feel free to use and distribute this file, but please report me if
!  you intend to change it (in better), so we can upgrade together
!
!  Released under GPL 2.0, or any later version at your will
!
!-----------------------------------------------------------------------------
!
! This files defines:
! - Class Spell: the basic spell class, whose objects can be cast etc.
! - Class AggressiveSpell: a subclass of the former, which deals damage
! - Class Memory: adding this to a player will make him a mage! (*)
! - Class Scroll: powerfull Spell source, or spell "reserve"
! - Class Spellbook: only source for spells
! - Standard spells of various form.
! (*) Your mage could be turned in something different than the player
! object; if you make him a Mage by subclassing the Player object, he
! will not retain his powers if it's turned in a frog, nor if it transform
! itself in a mighty dragon. Using a "carried" object, you are free to
! "move" it's powers along with the transformations.
! See "Damage.h" to understand some offensive spell priciples.
!-----------------------------------------------------------------------------

! only when first included
Ifndef Spell;

! Damage system needed.
Ifndef damage;
   Include "MFS_damage.h";
Endif;
! Pronoun system needed
Ifndef pron;
   Include "MFS_pronouns.h";
Endif;

! We need knowledge
Include "MFS_Knowledge.h";

!---------------------------------------------
! MFS Magic system constants
!
Constant MFS_TIMED 200;		! timed_spells. You have to concentrate.
Constant MFS_INSTANT 201;	! spells are fired immediately
Constant MFS_NEOPHITE 1;	! An apprentice wizzard
Constant MFS_WARLOCK 999;	! Can Cast everything
Constant MFS_VERBOSE 201;	! Verbose descriction of spells in memory
Constant MFS_STRICT 202;	! Describe only name and the rest
Constant MFS_BOOK_PRIORITY 3; !Priority for books && magical devices
Constant MFS_SCROLL_PRIORITY 2; !Priority for SCROLL
Constant MFS_MEM_PRIORITY 1; !Priority for memory
Constant MFS_NO_MULTIPLE 301; ! Not allow multiple learning
Constant MFS_ALLOW_MULTIPLE 302; ! Allow multiple learning
Constant MFS_FORGET 401; ! Too Much Spells - forget one
Constant MFS_FORBID 402; ! Too Much Spells - don't allow
Constant MFS_SINGLETARGET 501;	! for attack spells, fire_to_all hits 1 target
Constant MFS_MULTIPLETARGET 502; ! for attack spells, fire_to_all hits all targets
Constant MFS_SPELLVHI 602;		! high verbosity for spell stats
Constant MFS_SPELLVMED 601;	! medium verbosity for spell stats
Constant MFS_SPELLVLO 600;		! low verbosity for spell stats



!-------------------------------------------------------------
! Spell Effects (library default)
! affect property will be called with an effect of the
! following constants.

Constant SPELL_RECHARGE		1;
Constant SPELL_UNCURSE 		2;
Constant SPELL_REVEAL 		3;
Constant SPELL_DISPELL		4;
Constant SPELL_INVISIBLE	5;
Constant SPELL_PROTECT		6;
Constant SPELL_DESTROY		7;
Constant SPELL_REPAIR		8;
Constant SPELL_LOCK		 	9;
Constant SPELL_OPEN			10;
Constant SPELL_USER			19;

!----------------------------------------------------
! Declare alliance if combact system is not included
!
Ifndef alliance;
   Property alliance;
Endif;

!----------------------------------------------------
! Objects willing to give some kind of bonus for the
! spellcasters must provide giving_bonus
!
ifndef giving_bonus;
   Property giving_bonus;
endif;

!--------------------------------------------------------------
! Objects willing to receive spell effects must implement this
!
Property additive affect;

!----------------------------------------------------------
! This attribute will be given to the spells or other
! spell-like things if "they are free"; that is, they are
! always available in memory, and they do not inflict any
! form of cost (like the gnusto infocom spell).
!
Attribute for_free;

!----------------------------------------------------------
! Spell verbosity level. Change this to change the amount
! of informations displayed to the player when examinating
! spells.
!
Global SpellVerbosity = MFS_SPELLVHI;

!----------------------------------------------------------
! Grammar fake action to cast spells directly
!
Fake_action CastAt;
Fake_action FireSpellAt;

!-------------------------------------------------------
! Base function to find out the Memory Object. It's slow,
! so use it wisely
!
[GetMemory a;
   if ( a provides memory) return a.memory;

   if (a == player)
      print "Questa azione richiede che tu sia un VERO mago,
         ed in questo momento non lo sei^";
   return nothing;
];

!-------------------------------------------------------
! Helper function to see if a target can be affected by
! a spell, and then to affect it
!
[AffectTarget tg eff;
   if (~~ tg provides affect) rfalse;
   return tg.affect(eff);
];

!-------------------------------------------------------
! Helper function to print spell names, taking care of
! unknown spells
!
[SpellName sp;
   if (sp has known_about)
      print "l'incantesimo ", (name) sp;
   else {
      print (string) random (
         "uno sconosciuto",
         "uno strano",
         "un ignoto");
      print " incantesimo";
   }
];

[OfTheSpellName sp;
   if (sp has known_about)
      print "dell'incantesimo ", (name) sp;
   else {
      print (string) random (
         "di uno sconosciuto",
         "di uno strano",
         "di un ignoto");
      print " incantesimo";
   }
];

!-----------------------------------------------------------------------
!                          BaseMemory Class
!
! Memory class is the driver of the spellcasting system. It's capable
! of handle all the spell casting abilities and typologies. Simply, give
! your Player object a Memory object, and he will become a mage. You can
! have more than one Memory object, so you can give new abilities and
! powers easily to your mage, (or degrade him).
! The system supports a lot of spell casting styles:
! - Time delayed casting (cast, wait and fire) or immediate casting.
! - Target driven spells or generalized (targetless) spells.
! - Fatigue generating spells, Magic point consuming spells or "launch
!		and forget" spells (with magic books).
! The three spellcasting system are handled by different subclasses:
! - cast & forget (AD&D or INFOCOM adventures): CFMemory
! - Fatigue (damage) generating spells: FGMemory
! - Magic Points Consuming spells: MPMemory
! To set an appropriate subsystem of spellcasting, simply use the
! corresponding subclass.
! Note that BaseMemory is an "Abstract" class, that souldn't be
! used to instantiate an object.
!
Class BaseMemory with
   owner 0,
   room 0,
   cost_bonus 0,			! if you are a powerful mage...
   time_bonus 0,			!

   timeLeft 0,			! Time left before the spells goes off
   spellToFire,		! Spell to be casted
   target,				! ... guess!

   mage_power MFS_NEOPHITE, !Level of the mage
   timed_spells MFS_INSTANT,!Have you to wait a proper time?

   set [ o;
      if (o == nothing)
         self.owner = nothing;
      else {
         o.memory = self;
         self.owner = o;
      }
   ],

   ! Tells if there is a valid "casting source" in range
   cast_from[ sp obj bestobj bestpri pri;
      !magical ojects providing spells
      objectloop(obj in self.owner
         && obj provides can_cast ) {
            pri = obj.can_cast(sp);
            if (pri >= bestpri) {
               bestpri = pri;
               bestobj = obj;
            }
      }
      if (self.can_cast(sp) > bestpri) {
         bestobj = self;
         bestpri = MFS_MEM_PRIORITY;
      }

      if (bestpri == 0) return nothing;
      return bestobj;
   ],
   !-------------------------------
   ! And now, the casting routines!

   ! spell provider routine.
   cast[sp tr rm time;
      self.spellToFire = sp;
      self.room = rm;
      self.target = tr;
      ! find the global bonuses
      ! timeleft is reduced by one immediatly, so we
      ! have to add one to put things right.
      time = sp.casting_time - self.time_bonus -
         GlobalValue(self.owner,time_bonus,giving_bonus);

      if(time <=0 || self.timed_spells==MFS_INSTANT)
         self.fire();
      else {
         give self general;
         self.timeleft = time+1;
         StartDaemon(self);
      }
   ],

   !launch a readied spell
   fire [;
      self.spellToFire.room = self.room;
      self.spellToFire.caster = self.owner;
      if (self.target ~= 0) {
         if ( TestScope( self.target, self.owner )) {
            <FireSpell self.spellToFire self.target>;
            if (self.spellToFire hasnt for_free)
               self.inflict_cost(self.spellToFire);
         }
         else
            self.display(1);
      }
      else {
         <FireSpell self.spellToFire>;

         if (self.spellToFire hasnt for_free)
            self.inflict_cost(self.spellToFire);
      }
      self.spellToFire = 0;
   ],

   daemon [;
      self.timeLeft--;
      if (self.timeLeft <= 0) {
         self.fire();
         stopDaemon(self);
         return;
      }

      if (self has general) {
         give self ~general;
         self.display(0);
         return;
      }

      switch (self.timeLeft) {
         3: self.display(2); return;
         2: self.display(3); return;
         1: self.display(4); return;
         default:
            self.display(5);
      }
   ];

!-------------------------------------------------
! Memory objects for NPCs
!
Class NPC_Memory class BaseMemory
private
   pick [ num i;
      while (i < self.#content/2) {
         if (self.&content-->i ~=0) num--;
         if (num == 0) return self.&content-->i;
         i ++;
      }
      return 0;
   ],

with
   emulate MPmemory,
   mpoints,
   content,

   start_cast [sp tr rm src;
      self.room = rm;
      ! can I cast the spell?
      src = self.cast_from(sp);
      if (src == nothing) {
         rfalse;
      }

      ! if all the above condition are met, the spell
      ! provider is charged to launch it (according with
      ! the spell provider cost scheme).
      src.cast(sp,tr,rm,self.owner);
      rtrue;
   ],

   ! Signal that memory class can cast!
   can_cast [sp i;
      for( i = 0: i < self.#content/2: i++)
         if (sp == self.&content -->i)
            if (self.affordable(sp))
               return MFS_MEM_PRIORITY;
      return 0;
   ],

   inflict_cost[sp scost i;
      if (self.emulate == CFMemory) {
         for( i = 0: i < self.#content/2: i++)
            if (sp == self.&content -->i) {
               self.&content-->i = 0;
               return;
            }
      }

      scost = sp.cost - self.cost_bonus -
         GlobalValue(self.owner,cost_bonus,giving_bonus);
      if (scost <= 0) return;
      if (self.emulate == MPMemory)
         self.mpoints = self.mpoints -scost;
      else
         if (self.owner provides damage)
            self.owner.damage(scost, DMG_FATIGUE, self);
   ],

   affordable [sp scost;
      if (sp has for_free) rtrue;
      scost = sp.cost - self.cost_bonus - GlobalValue(self.owner,cost_bonus,giving_bonus);
      switch (self.emulate) {
         CFMemory: rtrue;
         MPMemory:
            if (scost <= self.mpoints) return true;
            rfalse;
         FGMemory:
            if (self.owner provides damage
               && self.owner provides health) {
               if (self.owner.health > scost)
                  rtrue;
               rfalse;
            }
            rtrue;
      }
   ],

   learn_spell[sp i;
      for(i = 0: i< self.#content/2: i++) {
         if (self.&content-->i == 0) {
            self.&content-->i = sp;
            rtrue;
         }
      }
      rfalse;
   ],

   count_spells [i num;
      for(i = 0: i< self.#content/2: i++)
         if (self.&content-->i ~= 0)
            num++;
      return num;
   ],

   pick_spell [num i s sp;
      num = self.count_spells();
      if (num == 0) return 0;
      s = random(num);
      sp = self.pick(s);
      if (sp == nothing) return nothing;

      ! verify if that spell can be cast.
      if(self.cast_from(sp) ~= nothing) return sp;

      ! else, enters this loop to find a suitable
      ! spell
      i = s+1;
      if ( i > num) i = 1;
      while (i ~= s) {
         sp = self.pick(i);
         if (self.cast_from(sp)) return sp;
         i = i + 1;
         if ( i > num) i = 1;
      }
      return 0;
   ],

   display[num;
      if (~~ TestScope(self.owner)) rfalse;
      switch(num) {
         0: print (cyouthe) self.owner, " intess";
            if ( self.owner == player ) print "i";
            else print "e ";
            print (SpellName) self.spellToFire, ".^"; return;
         1: 	"Il bersaglio designato ",
            (artda) self.owner, " non @`e pi@`u qui!";
         2:	print (cyouthe) self.owner, " continu";
            if ( self.owner == player ) print "i";
            else print "a";
            print " ad intessere ";
            SpellName(self.spellToFire);  ".";
         3: print (cyouthe) self.owner, " inizi";
            if ( self.owner ~= player ) print "a";
            print " ad intessere ";
            SpellName(self.spellToFire); "!";
         4: print (The) self.owner, " articol";
            if ( self.owner == player ) print "i";
            else print "a";
            print " le parole finali ",
            (OfTheSpellName)self.spellToFire,"!^";

            print  "Le dita ", (artdi) self.owner," crepitano di un potere che
            @`e pronto per scatenarsi!";
         5: print (cyouthe) self.owner, " ",(isorare) self.owner,
            " assorbit", (gan) self.owner, " dal lungo processo per lanciare ";
            SpellName(self.spellToFire);
            print ".";
      }
   ];

!------------------------------------------------
! Memory object for player
!
Class PC_Memory class BaseMemory
with
   known_spells 0,			 ! Nothing known
   memorized_spells 0,		 ! memorizing a spell you know twice...

   verbosity MFS_VERBOSE,	 ! We want to show all of the spells when listed
   cost_name "costo",		 ! The name of "mana" in your system
   multiple_spells MFS_NO_MULTIPLE,
   over_max MFS_FORBID,
   mem_max 0,				! any number of spells

   free_learn "Conosci gi@`a quest'incantesimo perfettamente.",
   no_spell_string	"Non conosci nessun incantesimo!",
   spell_string "Conosci i seguenti incantesimi:^", ! additional crlf
   low_mem_msg "La tua mente non @`e allenata per sopportare lo sforzo di
      apprendere cos@`i tanti incantesimi!",

   !---------------------------------
   ! General magic management section
   ! Learning, forgetting, listing and scribing.

   ! show what you know
   describe_contents
   [ i spell;
      i = 0;
      objectloop( spell in self && spell has for_free) {
         if (i == 0) print (The) spell;
         else print (the) spell;
         print ", ";
         i ++;
      }
      if (i > 0) {
         if (i == 1) print "@`e il solo incantesimo";
         else print "sono i soli incantesimi ";
         print "che sembra che tu abbia imparato davvero bene dopo tanti anni
         di studio, e rimangono tuoi per sempre. A parte questo,";
      }

      if (self.known_spells == 0)  {
         print "^",(string) self.no_spell_string,"^";
         rfalse;
      }

      print "^",(string) self.spell_string,"^";

      objectloop (i in self && i hasnt for_free) {
         i.specification(); !list all spells in myself
         if (self.verbosity == MFS_VERBOSE) {
            print "     Ha potere: ",i.power, " - ",
               (string) self.cost_name ,": ", i.cost;
            if (self.timed_spells == MFS_TIMED)
                  print " - Tempo di lancio: ",
                  i.casting_time;
            print ".^";
         }
      }
   ],

   ! Tells if there is a valid "learning source" in range
   learn_from[ sp obj bestobj bestpri pri;
      bestobj = nothing;
      bestpri = 0;

      !magical ojects providing learning
      objectloop(obj in player
         && obj provides can_learn ) {
            pri = obj.can_learn(sp);
            if (pri > bestpri) {
               bestpri = pri;
               bestobj = obj;
            }
      }
      return bestobj;
   ],

   ! message if you don't have a learning source in range
   cant_learn [;
      "Non c@`e nessun libro, pergamena o altro oggetto magico dal quale puoi imparare
      un incantesimo.^";
   ],

   ! message if you don't have a casting source in range
   ! Note: this routine is good for the three standard magic systems
   unable_to_cast [;
      print "In questo momento, non hai questo incantesimo memorizzato, o scritto su
      di una pergamena pronta per essere usata.^";
   ],

   ! To learn a spell, BaseMemory class puts the spell in itself;
   ! NEVER add or remove spells direclty, but always use this property

   learn[ sp src forget o index;

      forget = false;
      ! let's find a valid learning source
      src = self.learn_from(sp);
      if (src == nothing) {
         self.cant_learn();
         rfalse;
      }

      ! if you don't use spells like bullets to shot, you don't have
      ! to learn them twice.
      if (sp in self && self.multiple_spells == MFS_NO_MULTIPLE) {
         if ( sp has for_free)
            print free_learn,"^";
         else
            print "Conosci gi@`a questo incantesimo.^";
         rfalse;
      }

      ! tryed with a too powerful spell!
      if (self.not_power_enough(sp)) rfalse;

      ! too may spells ?
      if (self.mem_max > 0 && self.memorized_spells  >= self.mem_max)
         if (self.over_max == MFS_FORGET) {
            ! forget one random spell
            forget = true;
            ! forget a random spell...
            index = random(self.memorized_spells);
            objectloop(o in self) {
               index --;
               if (index <= 1) {
                  self.forget_spell(o);
                  break;
               }
            }
         }
         else { ! do not allow ...
            print "^",(string) self.low_mem_msg,"^";
            rfalse;
      }

      ! verifies again if the spell is already known
      if (sp notin self) {
         print "Le parole dell'incantesimo sono adesso fissate nella tua mente!^Ora ";
         sp.word_name();
         print " @`e pronto per essere utilizzato.^";
      }
      else {
         print "Con un grande sforzo, memorizzi ";
         sp.word_name();
         print " una seconda volta.^";
      }
      self.know_spell(sp);

      ! if I have to forget one...
      if (forget) {
         print "^Come inizi a studiare l'incantesimo, la tua mente diventa
         confusa e distratta. C'@`e semplicemente troppa confusione nella
         tua testa.^Quando finisci, percepisci che qualche altro
         incantesimo studiato in precedenza @`e stato dimenticato.^";
      }
      ! signal to the spell source that the spell has been
      ! learned successfully
      src.learn(sp);
         rtrue;
   ],

   !subclasses should overload this to see if there are others limitation
   !than maximum power: maximum spells in memory, or a filter for certain
   !spells etc.
   not_power_enough [sp;
      if (sp.power > self.mage_power +
         GlobalValue( self.owner, power_bonus, giving_bonus)  ){
         print "^... Ma l'incantesimo  troppo potente per te!
         la tua mente  tempestata di parole inintellegibili che non riesci
         a comprendere...^
         e di conseguenza, non potrai ricordare questo incantesimo.^";
         rtrue;
      }
      rfalse;
   ],

   ! to assign a spell without having the player to notice it.
   know_spell[ sp;
      if (sp has for_free) {
         give sp known_about;
         move sp to self;
         rtrue;
      }

      if (sp notin self) {
         move sp to self;
         give sp known_about;
         self.known_spells++;
      }
      else
         sp.commited++;

      self.memorized_spells++;
   ],

   ! and if something slips from memory
   ! (used by class CFMemory) - it cost almost nothing ,and it useful
   ! if you want somewhat erase a spell form a mage...
   forget_spell[sp;
      if (sp notin self) rfalse;

      if (sp.commited > 0)
         sp.commited--;
      else {
         self.known_spells--;
         remove sp;
      }
      self.memorized_spells--;
      rtrue;
   ],

   ! to transfers spells to other memories,
   transfer [mem obj;
      if (mem ofclass PC_Memory) {
         while (children(self) > 0) {
            obj = child(self);
            move obj to mem;
         }
         mem.mage_power = self.mage_power;
         mem.cost_bonus = self.cost_bonus;
         mem.time_bonus = self.time_bonus;
         mem.known_spells = self.known_spells;
         mem.memorized_spells = self.memorized_spells;
         mem.owner = self.owner;

         self.known_spells = 0;
         self.memorized_spells = 0;
         self.owner = 0;
         rtrue;
      }
      rfalse;
   ],

   !=========================================
   ! casting source routines
   !

   !.. if you cast a spell: start_cast(Spell, Target, room) ...
   start_cast [sp tr rm src;

      self.room = rm;
      ! can I cast the spell?
      src = self.cast_from(sp);
      if (src == nothing) {
         self.unable_to_cast();
         rfalse;
      }

      ! is the spellcasting right?
      if (tr == nothing && ~~(sp provides fire_to_all)) {
         print "La tua profonda conoscenza della magia ti dice che questo
         incantesimo deve avere un bersaglio.^";
         rfalse;
      }
      if (tr ~= nothing && ~~(sp provides fire)) {
         print "La tua conoscenza della magia di suggerisce che questo incantesimo
         non necessita un incantesimo.^";
         rfalse;
      }

      ! if all the above condition are met, the spell
      ! provider is charged to launch it (according with
      ! the spell provider cost scheme).
      src.cast(sp,tr,rm,self.owner);
      rtrue;
   ],

   ! Signal that memory class can cast!
   can_cast [sp;
      if (sp in self) return MFS_MEM_PRIORITY;
      return 0;
   ],

   ! subclasses may overload this, if you want your player to play
   ! the cost of an interrupted spell aniway.
   interrupted [;
      print "Hai perso la concentrazione, e non puoi portare a termine l'incantesimo!^";
      self.timeleft=0;	! deactivates waiting routine
      self.spellToFire = 0;
      StopDaemon(self);
   ],

   display[num;
         switch(num) {
            0:	print "Inizi a pronunciare il complesso incantesimo ";
               self.spellToFire.short_name();
               print ".^";
            return;
            1: 	"Il bersaglio che hai indicato non si trova pi@`u qui!";
            2: 	print "Continui ad intessere il complesso incantesimo ";
               self.spellToFire.short_name();
               ".";
            3:  print "Inizi a pronunciare le parole finali del ";
               self.spellToFire.short_name();
               "!";
            4: 	print "L'ultimo gesto del ";
               self.spellToFire.short_name();
               " @`e ultimato!^Il potere inizia a crepitare fra le tue dita,
               pronto per essere sprigionato!";
            5: print "Continui a elaborare il complesso ";
               self.spellToFire.short_name();
               ".";
         }
      ];

!------------------------------------------------------
! class to handle the MAGIC POINT spell casting system
!
class MPMemory class PC_Memory with
   mpoints 10,				! a standard dotation of magic points
   max_mpoints 10,
   cost_name "Magic Points", ! The name of "mana" in your system

   ! overload to display the magic points status; you can overload
   ! again this to fit the name of magic energy in your gaming system.
   describe_contents [;
      self.PC_Memory::describe_contents();
      print "^In questo momento possiedi ", self.mpoints,
         " su un totale di ", self.max_mpoints,
         " ",(string) self.cost_name,".";
   ],

   ! message if you don't have a learning source in range
   cant_learn [;
      print "Sfortunatamente, non possiedi un libro o una pergamena dai quali apprendere
      questo incantesimo.^";
   ],

   ! overload to see if it can be cast	(enough MP).
   start_cast [ sp tg rm rc;
      if (sp in self) {
         rc = sp.cost - self.cost_bonus - GlobalValue(self.owner,cost_bonus,giving_bonus);
         if (rc < 0) rc = 0;
         if (rc > self.mpoints)
            "In questo momento, il tuo potere @`e troppo debole per lanciare questo incantesimo.^";
      }
      self.PC_Memory::start_cast(sp,tg, rm);
   ],

   inflict_cost [sp rc;
      rc = sp.cost - self.cost_bonus - GlobalValue(self.owner,cost_bonus,giving_bonus);
      if (rc < 0) rc = 0;
      self.mpoints = self.mpoints - rc;
   ];

!------------------------------------------------------
! class to handle the FATIGUE GENERATING spell casting system
!
class FGMemory class PC_Memory
   with
   cost_name "Fatica",

   ! message if you don't have a learning source in range
   cant_learn [;
      print "Sfortunatamente, non possiedi un libro o una pergamena dai quali apprendere
      questo incantesimo.^";
   ],
   inflict_cost [sp rc;
      if (player provides damage) {
         rc = sp.cost - self.cost_bonus - GlobalValue(self.owner,cost_bonus,giving_bonus);
         if (rc < 0) rc = 0;
         player.damage(rc,DMG_FATIGUE, self);
      }
   ];


!------------------------------------------------------
! class to handle the CAST and FORGET spell casting system
! If you use this, you'll have to give a spell book to the
! player, or refresh his memory in some way...
!
class CFMemory class PC_Memory
   with
   cost_name nothing,
   mem_max	4,				! maximum number of spells that can be held
   multiple_spells MFS_ALLOW_MULTIPLE,
   over_max MFS_FORGET,

   no_spell_string 
      "Attualmente, non hai appreso ancora nessun incantesimo incantesimo.",
   spell_string "Hai appreso i seguenti incantesimi:^",

   inflict_cost [sp;
      self.forget_spell(sp);
   ];

!--------------------------------------------------------------
! This is the Spell Class.
! Spells can be cast by the player saying
! > CAST XYZZY [ON (THE) TARGET]
! or
! > XYZZY [(THE) TARGET]
! The latter method requires the "DirectCastSub" parse routine.
!
! The system supports different spellsystem styles, that can be selected
! from a Memory class derived object.
! used not in listing, but in description (examine) of spells


class Spell
class Knowledge
with
   name "spell",

   ! particular definitions
   power 1,        ! limits the ability of a mage to learn (and use) it
   casting_time 1,    ! Casting time needed (if MFS_TIMED)
   cost 1,            ! Cost of the spell (for MP or FATIGUE systems)
   commited 0,		  ! how many times the mage have studied it
               ! for cast and forget system
   room,
   caster,

   ! the simple name of the spell
   word_name
   [;  print (address) (self.&name)-->0; rtrue;
   ],

   ! if the player see it, he can "call it"
   short_name
   [;  self.word_name(); give self known_about; rtrue;
   ],

   ! listing description
   specification [;
      print "Il ";
      style bold;
      self.short_name();
      style roman;
      if (self.commited > 0 && sender ofclass BaseMemory) print " (memorizato ", self.commited+1,
      " volte)";
      print ": ", (string) self.purpose,".^";;

   ],

   explain[ memory;
      print "Scopo: ",(string) self.purpose,".^";
      if (self provides description &&
         SpellVerbosity > MFS_SPELLVLO) {
         if (metaclass(self.description) == Routine)
               self.description();
         else if (self.description ~= 0)
               print (string) self.description,"^";
      }

      if (SpellVerbosity > MFS_SPELLVMED) {
         memory = GetMemory(player);
         print "^Potere: ", self.power;
         if (cost_name ~= nothing)
            print "^",(string)memory.cost_name,
               ": ",self.cost;
         if (memory.timed_spells == MFS_TIMED)
            print "^Tempo di lancio: ",self.casting_time;
         print "^";
      }
      rtrue;
   ];

!-----------------------------------------------------------------------
! Damaging a target with a spell
! Since all damaging (or healing) spell are similar in effect,
! the 'fire' and 'fire_to_all' properties are standardized.
!
class AttackSpell class Spell
with
   min_damage 5,
   max_damage 10,
   damage_type DMG_MAGIC,
   target_type MFS_SINGLETARGET,	! single or multiple

   ! Description of the effect: since it's standardized, you can
   ! change only a fiew string (moslty prelude_string) instead of
   ! rewriting a new routine for every aggressive spell description.

   print_effect[stage target room;
      if (target ~= nothing) {
         if (~~TestScope(target)) return;
      }
      else if (location ~= room) return;
      self.effect(stage,target);
   ],

   effect[stage target;
      print "L'incantesimo aggressivo @`e lanciato";
      if (target ~= nothing) {
         print (artsu) target, ".";
      }

      switch(stage) {
         3: "...^ Ma non ci sono bersagli nei pressi.";
         2: ", e colpisce!";
         1: "e colpisce, ma sembra non causare alcun danno ",(atytpacc) target;
         0: " ma l'incantesimo manca ",(ytpacc) target,"!";
      }
   ],

   fire [target;
      if (target provides damage) {
         if (target.damage(0,self.damage_type, self.caster)){
            self.print_effect(2,target);
            target.damage(self.min_damage+
               random(self.max_damage-self.min_damage),
               self.damage_type,
               self.caster);
         }
         else self.print_effect(1,target);
      }
      else self.print_effect(0,target);

   ],

   fire_to_all [ obj fired all;
      !finds the caster and it's alliance

      fired = false;
      all = -1;

      if (self.caster ~= nothing && self.caster provides alliance)
         all = self.caster.alliance;

      objectloop(obj in self.room) {
         if (obj provides damage) {
            if ( obj == self.caster || (
                  obj provides alliance &&
                  obj.alliance == all))
               continue;
            else {
            fired = true;
            <FireSpell self obj>;
            if (self.target_type ==
                  MFS_SINGLETARGET) return;
            }
         }
      }
      if (~~fired) self.print_effect(3,0,self.room);
   ];



!--------------------------------------------------------------------
! Manages the unfold verb
!
[UnfoldSub;
   if (noun ofclass Scroll) {
      if (BeforeRoutines()) rtrue;
      noun.unfold();
   }
   else
      "Forse ti potrebbe venire in mente un modo migliore di usarl", (gan) noun,".";
   return AfterRoutines();
];

!-----------------------------------------------------------
! Manage "memorize verb"
!
[MemorizeSub memory;
   if ( (memory = GetMemory(actor)) == nothing)rtrue;
   memory.learn(noun);
   return AfterRoutines();
];

!--------------------------------------------------------------
! Manages "scribe" verb
!
[ScribeSub memory;
   if ( (memory = GetMemory(actor)) == nothing) return;

   if ( noun notin memory )
      "Devi avere l'incantesimo in memoria per poterlo scrivere su di una pergamena.";

   if (second ~= nothing && second provides scribe_spell) {
      second.scribe_spell(noun);
   }
   else
      "Puoi scrivere incatesimi solo su pergamene o altri oggetti magici scrivibili.";
   return AfterRoutines();
];

!-------------------------------------------------------------------
! Tells if a spell is in known, and therefore referenced by player
!
[KnownSpells obj;
   switch(scope_stage) {
      1: rfalse;
      2: objectloop(obj ofclass Spell && obj has known_about)
            PlaceInScope(obj);
         rtrue;
      3: "E che cosa sarebbe? Un incantesimo? Una ninna nanna?";
   }
];

!---------------------------------------------------------
! Routine that responds to the "spells" meta verb command
!
[ SpellsSub memory;
   memory = GetMemory(actor);
   if (memory == nothing) return;

   memory.describe_contents();
];

!----------------------------------------------------
! Tryng to cast a spell (manages "cast" verb)
!
[ CastSub o;
   ! if I am not a mage, I should not be playing with magic...
   o = GetMemory(actor);
   if (o == nothing)	return;
   ! let's (try to) launch the spell
   !if (the_spell ~= nothing ) {
   !  second = noun;
   !   noun = the_spell;
   !   the_spell = nothing;
   !}

   if (second ~= nothing) {
      action = ##CastAt;
      if ( RunRoutines( second, before ) ) {
         action = ##Cast;
         rtrue;
      }
      action = ##Cast;
   }
   o.start_cast(noun,second,location);

   if (second ~= nothing) {
      action = ##CastAt;
      if ( RunRoutines( second, after ) ) {
         action = ##Cast;
         rtrue;
      }
      action = ##Cast;
   }

   return AfterRoutines();
];

[ FireSpellSub;

   if (second ~= nothing) {
      action = ##FireSpellAt;
      if ( RunRoutines( second, before ) ) {
         action = ##FireSpell;
         rtrue;
      }
      action = ##FireSpell;
      noun.fire( second );
   }
   else
      noun.fire_to_all();

   if (second ~= nothing) {
      action = ##FireSpellAt;
      if ( RunRoutines( second, after ) ) {
         action = ##FireSpell;
         rtrue;
      }
      action = ##FireSpell;
   }
   
   return AfterRoutines();
];

!-------------------------------------------------------
! Convert a spell verb into the cast action
!
[CastDirectSub;
   second = noun;
   noun = the_spell;

   if ( second ~= nothing )
      <<Cast noun second>>;
   else
      <<Cast noun>>;
];

!------------------------------------------
! Manages the "prepare" verb.
!
[PrepareSub;
   if (noun provides prepare) {
   noun.prepare();
   }
   else
      "Non vedo come potresti!";
   return AfterRoutines();
];

[Prepareable;
   return (noun provides prepare );
];

!--------------------------------------------------------
! Manages the "repose" verb
!
[ReposeSub;
   if (noun provides repose) {
      noun.repose();
   }
   else
      "@`E gi@`a ripost", (gan) noun, " fra le tue cose!";
   return AfterRoutines();
];

!----------------------------------------------------
! recognise a spell in the form "spell_name at..."
!

Endif; ! ifndef Spell

! end of MFS_Magic.h

