!-----------------------------------------------------------------------------
! Magic And Fight System --- TopicParse, a general topic parsing routine
! For INFORM compiler and Z-Machine text based Adventure games
!-----------------------------------------------------------------------------
!  Author: Giancarlo Niccolai
!  Version: 1.3
!  Last Revised: 2002-11-24
!  Contact author at: giancarlo@niccolai.ws
!  Feel free to use and distribute this file, but please report me if
!  you intend to change it (in better), so we can upgrade together
!
!  Released under GPL 2.0, or any later version at your will
!
!-----------------------------------------------------------------------------
! Returns the topic of the conversation or of the consultation the player is
! willing to have. The topic returned can be a 'string' or an object, if the
! parser is able to find an object refereced by the player.
!
! Useage example:
!
! Object BlackRobe "Black Robe" with name "black" "robe";
! ..
! Object seller "seller" with
! ...
!   life [;
!      switch( MFS_TopicParse( MFS_Topic_All ) ) {
!        black_robe: "~This robe is made of pure silk!~";
!        'weather': "~A really rainy day, sir!~";
!        nothing: "What are you talking about?";
!      }
!   ]...
!
! You can control which object are referenced by the player using an optional parameter
!
! MFS_TopicParse( sw )
!   sw = 0 (default) Return only the first topic word
!   sw = MFS_Topic_All(1) Returns also objects;
!
! return :
!  nothing - topic in phrase is not in dictionary, or noun object can be confused
!  MFS_Undefined - More than one object matches: do not print a thing
!  Else it will return the object named by the player; if no object is named,
!  a word will be returned.
!
!-----------------------------------------------------

Ifndef MFS_TopicParse;

Ifndef MFS_Undefined;
 	Constant MFS_Undefined = -1;
Endif;

Include "MFS_BasicWords.h";


! [MFS_AskWhich match obj count cached word i;
! 
! 	!LanguageLM(46);
! 
! 	print "Quale intendi ";
! 	cached = 0;
! 	objectloop( obj ) {
! 		wn = consult_from;
! 		count = 0;
! 		while ( wn-consult_from  < consult_words ) {
! 			word = NextWord();
! 
! 			if ( MFS_Discard( word ) )
! 				continue;
! 
! 			for (i = 0: i < obj.#name/2: i++ ) {
! 				if ( word == obj.&name-->i ) {
! 					count ++;
! 					break;
! 				}
! 			}
! 		}
! 
! 		! Total match? no need to go beyond!
! 		if ( match == count ) {
! 			if ( cached ) {
! 				print ", ";
! 				print (the) cached;
! 			}
! 			cached = obj;
! 		}
! 	}
! 	print " o ", (the) cached,"?^";
! ];

![MFS_ScanObjects obj i word found match_found match;
!	found = nothing;
!	match_found = 0;
!
!	objectloop( obj ) {
!		wn = consult_from;
!		match = 0;
!
!		while ( wn-consult_from  < consult_words ) {
!			word = NextWord();
!
!			if ( MFS_Discard( word ) )
!				continue;
!
!			! some common check
!			if ( word == 'me' or 'myself' ) return player;
!
!			for (i = 0: i < obj.#name/2: i++ ) {
!				if ( word == obj.&name-->i ) {
!					match ++;
!					break;
!				}
!			}
!		}
!
!		! Total match? no need to go beyond!
!		if ( match && match == obj.#name/2 && wn-consult_from  == consult_words) {
!			return obj;
!		}
!
!		! ooops... an indecision!
!		if ( match && match == match_found ) {
!			MFS_AskWhich( match );
!			return MFS_Undefined;
!		}
!
!		if (match > match_found ) {
!			found = obj;
!			match_found = match;
!		}
!	}
!	return found;
!];


[MFS_TopicParse ;

! 	if (sw ~= MFS_Topic_All ) return candidate;
! 
! 	! Now let's find an object with name compatible
! 	wn = consult_from;
! 
! 	cando = MFS_ScanObjects();
! 
! 	if ( cando ~= REPARSE_CODE or nothing ) return cando;
	return MFS_GetWord();
];


Endif;

! end of MFS_TopicParse.h
